; screen.asm
; Screen rendering code
; -BGMODE specific handlers, including specialized handlers for:
;   layer palettes in Mode 0;
;   partial offset change in Mode 2;
; -Basic BG scanline renderers;
; -Line offset in tile tables.
;

;%define NO_OFFSET_CHANGE
;%define NO_OFFSET_CHANGE_DISABLE
;%define NO_16x16_TILES

%define SNEeSe_screen_asm

%include "misc.ni"
%include "sprites.ni"
%include "PPU.ni"
%include "screen.ni"

extern SCREEN_MODE_7
extern ScreenDestPtr
extern _Offset_Change_Disable

extern _TileCache2,_TileCache4,_TileCache8
extern Tile_Line_8_2,Tile_Line_8_4,Tile_Line_8_8
extern Tile_Line_16_2_Even,Tile_Line_16_4_Even
extern Tile_Line_16_2,Tile_Line_16_4,Tile_Line_16_8

extern TilesetAddress

section .text
EXPORT_C screen_text_start
section .data
EXPORT_C screen_data_start
section .bss
EXPORT_C screen_bss_start

section .data
ALIGND
%macro Generate_Tile_Offset_Table 1
 dd 16*64*(%1)
 dd 16*64*(%1)+1*8
 dd 16*64*(%1)+2*8
 dd 16*64*(%1)+3*8
 dd 16*64*(%1)+4*8
 dd 16*64*(%1)+5*8
 dd 16*64*(%1)+6*8
 dd 16*64*(%1)+7*8
%endmacro

EXPORT Tile_Offset_Table_64_8
Generate_Tile_Offset_Table 0
Generate_Tile_Offset_Table 1
Generate_Tile_Offset_Table 2
Generate_Tile_Offset_Table 3
Generate_Tile_Offset_Table 4
Generate_Tile_Offset_Table 5
Generate_Tile_Offset_Table 6
Generate_Tile_Offset_Table 7

EXPORT Screen_Mode
 dd SCREEN_MODE_0
 dd SCREEN_MODE_1
 dd SCREEN_MODE_2
 dd SCREEN_MODE_3
 dd SCREEN_MODE_4
 dd SCREEN_MODE_5
 dd SCREEN_MODE_6
 dd SCREEN_MODE_7

Pixel_Increment:dd 1,-1
First_Pixel:    dd 0,7

section .bss
ALIGNB
EXPORT LineAddress ,skipl   ; Address of tileset, + offset to line in tile
EXPORT LineAddressY,skipl   ; Same, for vertical flip
EXPORT TileMask    ,skipl   ; Tile address mask, for tileset wrap
Current_Line_Offset:skipl   ; Current render line # for offset change layers
                            ; Used for vertical mosaic effect

MapAddress_Current: skipl   ; Scroll-adjusted tile map address for offset
                            ;  change code when tile offset not changed
MapAddress_Base:    skipl   ; Base tile map address for offset change code
LineAddressOffset:  skipl   ; Same as LineAddress(Y), additional vars
LineAddressOffsetY: skipl   ;  for offset change code

; Used for offset change map addressing
EXPORT OffsetChangeMap_HOffset,skipl    ; BG3HOFS + BG3SC
EXPORT OffsetChangeMap_VOffset,skipl    ; BG3VOFS + BG3SC

EXPORT Render_Select,skipl  ; Base renderer
EXPORT Render_Mode  ,skipl  ; Mode renderer

EXPORT Tile_priority_opcode,skipb

section .text
ALIGNC
EXPORT Render_Layering_Option_0 ; sub-on-main
 mov al,[SCR_TM]    ; Get BG status for main screens
 call dword [Render_Mode]
 mov al,[SCR_TS]    ; Get BG status for sub screens
 jmp dword [Render_Mode]

ALIGNC
EXPORT Render_Layering_Option_1 ; main-on-sub
 mov al,[SCR_TS]    ; Get BG status for sub screens
 call dword [Render_Mode]
 mov al,[SCR_TM]    ; Get BG status for main screens
 jmp dword [Render_Mode]

ALIGNC
EXPORT Render_Layering_Option_2 ; main-with-sub
 mov al,[SCR_TM]    ; Get BG status for main/sub screens
 jmp dword [Render_Mode]

%macro SORT_SCREEN_LAYOUT 1
 mov al,[Redo_BGs]
 LOAD_BG_TABLE %1
 test al,1<<(%1-1)
 jz %%no_recalc

 call dword [BG_Redo+edx]

%%no_recalc:
 mov al,[TileHeight+edx]
 mov cl,[TileWidth+edx]
 mov [Tile_Height],al
 mov [Tile_Width],cl
 call Sort_Screen_Height
%endmacro

ALIGNC
EXPORT Sync_Render
 ; edi = number of lines to recache
 push eax
 push ebx
 push ecx
 push edx
 push ebp
 push esi
 push edi

extern BGMODE_Tile_Layer_Mask
 mov al,[SCR_TM]
 mov bl,[BGMODE_Tile_Layer_Mask]
 and [Redo_BGs],bl
 or al,[SCR_TS]
 and al,bl
 jz .no_tile_recache_needed ; no tile layers?

extern Tile_Recache_Set_End,Tile_Recache_Set_Begin,Recache_Tile_Set
 mov edi,[Tile_Recache_Set_End]
 inc edi
 js .no_tile_recache_needed ; No set to recache?
 push eax
 sub edi,[Tile_Recache_Set_Begin]
 call Recache_Tile_Set
 mov dword [Tile_Recache_Set_Begin],-2
 mov dword [Tile_Recache_Set_End],-2
 pop eax
.no_tile_recache_needed:

 test al,0x10
 jz .no_oam_recache_needed
 mov al,[Redo_OAM]
 test al,al
 jz .no_oam_recache_needed
 call Recache_OAM
.no_oam_recache_needed:

 ; Clear all lines we're drawing to
%if 0
 mov edi,[ScreenDestPtr]

 xor eax,eax
 mov ecx,[esp]
 mov ebp,256/32
 jmp .clear_loop
 
ALIGNC
.clear_loop:
 lea esi,[edi+16]
 mov [edi],eax
 mov [edi+4],eax
 mov [edi+8],eax
 mov [edi+12],eax
 mov [esi],eax
 lea edi,[esi+16]
 mov [esi+4],eax
 mov [esi+8],eax
 mov [esi+12],eax
 dec ebp
 jnz .clear_loop

 add edi,byte 64    ; Point screen to next line
 dec ecx
 mov ebp,256/32
 jnz .clear_loop
%endif

.next_line:
 inc dword [Current_Line_Render]

 mov ah,[_INIDISP]
 mov edi,[ScreenDestPtr]
 test ah,ah         ; Check for screen off
 js near .line_done

.render_line:
 CALCULATE_MOSAIC_YLINE

 SORT_SCREEN_LAYOUT 1
 SORT_SCREEN_LAYOUT 2
 SORT_SCREEN_LAYOUT 3
 SORT_SCREEN_LAYOUT 4
 mov byte [Redo_BGs],0

 call dword [Render_Select]
 mov edi,[ScreenDestPtr]
.line_done:
 add edi,256+64     ; Point screen to next line
 dec dword [esp]
 mov [ScreenDestPtr],edi
 jnz near .next_line
 pop edi
 pop esi
 pop ebp
 pop edx
 pop ecx
 pop ebx
 pop eax
 ret

; ~142 bytes
;%1 = Left,%2 = Right,%3 = ICount,%4 = IBands,%5 = OCount,%6 = OBands
%macro Recalc_Window_Bands 6
 mov al,[%1]
 mov dl,[%2]
 test dl,dl         ; 0 = 255 (right edge)
 jz %%one_inside
 cmp dl,al
 jbe %%full_outside ; if (Right < Left) full range outside window;
%%one_inside:
 mov [%4],al
 mov [%4+1],dl
 mov byte [%3],1    ; One band inside window (left,right)
 test al,al
 jnz %%not_flush_left   ; if (!Left) window flush left;
 test dl,dl
 jne %%flush_one_side   ; if (!Left && Right == 255) full range inside;
 ; Full range inside window
 mov byte [%5],0  ; No bands outside window
 jmp %%done
%%not_flush_left:
 ; Window not flush left (1 inside, 1 or 2 outside)
 test dl,dl
 je %%flush_one_side    ; if (Left && Right == 255) window flush right;
 ; Window not flush left or right (1 inside, 2 outside)
 ; Inside range is (left,right)
 ; Outside range 1 is (0,left-1)
 ; Outside range 2 is (right+1,255)
;dec eax                ; Right outside edge 1 = Left inside edge - 1
;inc edx                ; Left outside edge 2 = right inside edge + 1
 mov byte [%5],2        ; One band outside window (right+1,left-1)
 mov [%6+1],al
 mov byte [%6],0
 mov [%6+2],dl
 mov byte [%6+3],0
 jmp %%done
%%flush_one_side:
 ; Window flush left, not flush right (1 inside, 1 outside)
 ; Window flush right, not flush left (1 inside, 1 outside)
 ; Inside range is (left,right), outside range is (right+1,left-1)
;dec eax                 ; Right outside edge = Left inside edge - 1
;inc edx                 ; Left outside edge = right inside edge + 1
 mov [%6+1],al
 mov byte [%5],1         ; One band outside window (right+1,left-1)
 mov [%6],dl
 jmp %%done
%%full_outside:
 ; Full range outside window (0 inside, 1 outside)
 mov byte [%5],1         ; One band outside window
 mov dword [%6],0        ; Full range band
 mov byte [%3],0         ; No bands inside window
%%done:
%endmacro

ALIGNC
EXPORT Recalc_Window_Effects
 push eax
 push edx
 Recalc_Window_Bands _WH0, _WH1, _Win1_Count_In, _Win1_Bands_In, _Win1_Count_Out, _Win1_Bands_Out
 Recalc_Window_Bands _WH2, _WH3, _Win2_Count_In, _Win2_Bands_In, _Win2_Count_Out, _Win2_Bands_Out

 pop edx
 pop eax
 ret
; full | any == full
; full & any == any
; side-flush | same-flush == same-flush (longer)
; side-flush & same-flush == same-flush (shorter)
; side-flush ^ same-flush == 0 (if equal) or no-flush (difference)
; side-flush & other-flush == 0 (if no overlap) or no-flush (overlap)

; windows:
%if 0
TMW/TSW clipping is only done inside window areas defined by
 WH0-WH1 (window 1) and WH2-WH3 (window 2), when enabled
 (W12SEL/W34SEL/WOBJSEL odd bits).

Window areas can be inverted (W12SEL/W34SEL/WOBJSEL even bits).

When specified areas of windows 1/2 overlap, final window area is determined
 by specified logic (WBGLOG/WOBJLOG).

Color arithmetic is done inside the area of the color window.

w2  w1
 /\/\  /--+ enable
 ||||  |/-+ invert window area
 76543210
 \  /\  /
 BG2 BG1 - ($2123) W12SEL
 BG4 BG3 - ($2124) W34SEL
 COL OBJ - ($2125) WOBJSEL

 COL = Color window - related to CGWSEL ($2130)

WH0-WH1 Left and right position for window 1
WH2-WH3 Left and right position for window 2
 if (left > right) no window range

       /+-+ logic - 00 = or; 01 = and; 10 = xor; 11 = xnor
       ||
 76543210
 \/||||\/
BG4||||BG1 ($212A: WBGLOG)
   \/\/OBJ ($212B: WOBJLOG)
 BG3  BG2 ($212A: WBGLOG)
      COL ($212B: WOBJLOG)
 bits 4-7 are ignored in $212B: WOBJLOG

 bits 5-7 are ignored in $212C-212F (TM, TS, TMW, TSW)
 76543210
 xxx||||\-+ BG1
    |||\--+ BG2
    ||\---+ BG3
    |\----+ BG4
    \-----+ OBJ
 ($212C) TM specifies layers to be used as main screen
 ($212D) TS specifics layers to be used as sub screen, for screen arithmetic
 ($212E) TMW is mask to be applied with bitwise AND to TM inside window areas
 ($212F) TSW is mask to be applied with bitwise AND to TS inside window areas

 $2130 - CGWSEL
 76543210
 ||||xx|\-+ 1 = enable direct color mode (for BGMODEs 3,4,7)
 ||||  \--+ 0 = arithmetic with fixed color; 1 = arithmetic with screen
 ||\+-----+ sub screen normal display select: 00 = on; 01 = on inside
 \+-------+ main screen normal display select: 10 = on outside; 11 = off

 $2131 - CGADSUB

 76543210
 |||||||\-+ enable color arithmetic for BG1
 ||||||\--+ enable color arithmetic for BG2
 |||||\---+ enable color arithmetic for BG3
 ||||\----+ enable color arithmetic for BG4
 |||\-----+ enable color arithmetic for OBJ
 ||\------+ enable color arithmetic for back area
 |\-------+ 1 = halve-result of arithmetic (except for back area)
 \--------+ 0 = color addition; 1 = color subtraction
%endif

; esi is screen address, works cos we only plot until wraparound!
; ch contains the X counter
; cl contains the screen addition for palette offsetting (2bpl only)
; edi is the address to draw to..
; LineAddress contains the location for the SNES tile data
; LineAddress(Y) must be offset to the correct line for that row
; LineAddressY is used for Y-flip

%if 0
 Mode 2 is a special snes mode - It is known as offset change mode,
  basically the snes has the ability to change the horizontal and/or
  vertical offset of each column on the screen, a pig to emulate but
  here is the information:

  BG1 & BG2 are 16 colour planes (this much was simple to find out)
  BG3 - This does not exist but its address in VRAM is very important!

  What happens is the horizontal and verticle information is written to
  BG2 address's BG2+0 -  BG2+63, this gives 128 bytes (since VRAM is a
  word addressed system). BG2+0 -  BG2+31 is the address for changing
  the horizontal value BG2+32 - BG2+63 is the address for changing the
  vertical value

  There are 32 values per BG mode since there are 32 tiles horizontally across the screen! My
  best guess is this value is immune to scrolling (otherwise it would make more sense to
  give a 64 tile range in case of extra width modes).

  Ok, the only other thing you need to know is what do the values at address BG2+x do.

  Well its the same for horiz as vertical the values are encoded as :

  00           | 01          | 02           | 03          |.......| 62            | 63
  col 0 offset | col 0 flags | col 1 offset | col 1 flags |.......| col 31 offset | col 31 flags

  The flags as far as I can tell are :

   bit 7 6 5 4 3 2 1 0
       ? y x ? ? ? ? ?		where	y = affect bg1
    				x = affect bg0

    			The above information came from me Savoury SnaX

Addendum by TRAC (0.33 -> 0.34)
 BG modes 2/4/6 have offset change support.
 Offset change info is always stored at address set via BG3SC ($2109)?

 Offset change info is stored one word per (width-8) tile, which appears
  to have the format of:

  FEDCBA9786543210
  421xxxoooooooooo

  4 = reserved in modes 2/6 - vertical select in mode 4
  o = offset
  1 = enable for first layer
  2 = enable for second layer

 The (width-8) tile on the left edge of a layer cannot have its offset
  changed. It will always use standard H/V scroll.

 The offset change map is used for the remaining width of the screen.
  The maximum number of displayed tiles is 33, and the leftmost tile is
  excluded, hence there is data in the table for 32 tiles.

 There are one or two sets of data for the scanline. Each set is 32 words
  (one word per tile). In BG modes 2 and 6 there are two: one for
  horizontal, followed by one for vertical. In BG mode 4, there is one,
  shared between horizontal and vertical.

 Mode 4 stores only one word of offset change data per tile - bit 15
  of a word entry determines if it is used for changing horizontal or
  vertical offset (set for vertical).

 When offset change is enabled for a tile, it replaces the standard offset
  (BG#HOFS, BG#VOFS) value for that tile (horizontal offset replaces
  BG#HOFS, vertical offset + 1 replaces BG#VOFS).

 The BG3 scroll registers move the offset change map. The scrolling is
  limited to increments of 8.

 ; word offset in offset change map of first offset change entry
 OffsetChangeMap_X = BG3HOFS >> 3;
 ; word offset in row in offset change map to use
 OffsetChangeMap_Y = (BG3VOFS >> 3) << 5;

 Note that the offset change map does wrap rows back to themselves like
  any other layer.

%endif

ALIGNC
; -Tile on left edge of screen not affected by V-offset change
; -Offset change map is scrollable - always 8x8
EXPORT_C RENDER_8x8_C4_Offset
EXPORT RENDER_8x8_C4_Offset
%ifndef NO_OFFSET_CHANGE
%ifndef NO_OFFSET_CHANGE_DISABLE
 cmp byte [_Offset_Change_Disable],0    ; Hack to disable offset change
 jnz near RENDER_8x8_C4
%endif  ; !NO_OFFSET_CHANGE_DISABLE
%else   ; NO_OFFSET_CHANGE
 jmp RENDER_8x8_C4
%endif  ; NO_OFFSET_CHANGE

 mov cl,[OC_Flag+edx]
 SORT_TILES_8_TALL offset_change
 ; This sets up the screen vert 2!

 mov al,[Tile_priority_opcode]
 mov edi,[SetAddress+edx]
 mov [LINE_8x8_C4_Offset.priority],al
 add edi,[_TileCache4]
 mov [TilesetAddress],edi
 xor ebp,ebp
 test byte [BG3SC],1
 jz .no_wide_offset_map
 mov ebp,32*32*2

.no_wide_offset_map:
 mov [LINE_8x8_C4_Offset.Offset_RMapDifference],ebp

 push ecx
 ; Corrupts eax,ecx,ebp
 mov eax,[TLMapAddress+edx]
 mov edi,[BLMapAddress+edx]
 mov [LINE_8x8_C4_Offset.TMapAddress],eax
 sub edi,eax
 mov [LINE_8x8_C4_Offset.BMapDifference],edi

 mov ecx,[Current_Line_Offset]
 add ecx,[VScroll+edx]
 test ch,1
;global Tile_Height
;Tile_Height equ $-1    ; 1 = 8x8, 16x8, 2 = 16x16
 jz .current_line_in_screen_map_top
 add eax,edi
.current_line_in_screen_map_top:
 mov edi,[MapAddress_Base]
 add edi,eax
 mov [MapAddress_Current],edi
 pop ecx

 mov eax,[HScroll+edx]
 shr eax,3
 add eax,eax
 mov [LINE_8x8_C4_Offset.FirstTile],eax

 xor ebp,ebp
 test byte [BGSC+edx],1
 jz .no_wide_screen_map
 mov ebp,32*32*2
.no_wide_screen_map:
 mov [LINE_8x8_C4_Offset.RMapDifference],ebp

 ; cl contains bit for determining planes to affect

 ; BG3 scroll affects offset-change map address
 mov eax,[OffsetChangeMap_HOffset]
 mov ebx,[OffsetChangeMap_VOffset]  ; BG3VOFS adjusted
 sub eax,byte 2
 mov [LINE_8x8_C4_Offset.MapAddress_OffsetChange],ebx

 mov edi,[ScreenDestPtr]
 sub edi,[FirstPixel+edx]
 mov ch,[TileCount1+edx]
 add ch,[TileCount2+edx]
 xor esi,esi
 jmp LINE_8x8_C4_Offset.LeftEdge

ALIGNC
LINE_8x8_C4_Offset:
.next_tile:
 push ebx
 mov eax,ebx

 and ebx,byte 31*2  ; X offset wrap
 test eax,32*2
 jz .current_tile_in_offset_map_left
 add ebx,0x80000000
.Offset_RMapDifference equ $-4
.current_tile_in_offset_map_left:
 mov eax,[Current_Line_Offset]
 add ebx,0x80000000
.MapAddress_OffsetChange equ $-4

 push ecx
 mov dh,[1+ebx]     ; Horizontal offset change map
 test cl,dh
 jz .No_HChange
 mov dl,[ebx]
 mov ebp,edx
 shr ebp,3
 add ebp,ebp
 jmp .have_h_offset
.No_HChange:
 mov ebp,0x80000000
.FirstTile equ $-4
.have_h_offset:
 mov dh,[32*2+1+ebx]    ; Vertical offset change map
 test cl,dh
 jz .No_VChange
 mov dl,[32*2+ebx]
 add eax,edx
 jmp .calc_v_offset

.LeftEdge:
 push eax
 push ecx

 mov ebp,[.FirstTile]

.No_VChange:
 mov eax,[LineAddress]
 mov ecx,[MapAddress_Current]
 mov edx,[LineAddressY]
 jmp .have_v_offset

.calc_v_offset:
 test ah,1
 mov ebx,0x80000000
.TMapAddress equ $-4
 jz  .line_in_screen_map_top
 add ebx,0x80000000
.BMapDifference equ $-4
.line_in_screen_map_top:

 ; al contains real Y offset for next tile
 shl eax,3      ; Get screen offset
 mov ecx,0xF8*8
 and ecx,eax
 mov edx,8*7
 add ecx,ebx
 and eax,byte 0x07*8
 sub edx,eax
.have_v_offset:
 add ebp,esi
 test ebp,32*2
 jz  .tile_in_offset_map_left
 add ecx,0x80000000
.RMapDifference equ $-4
.tile_in_offset_map_left:
 and ebp,byte 31*2  ; X offset wrap
 add ecx,ebp        ; Combine X and Y offsets into tile map
 mov [LineAddressOffset],eax
 add esi,byte 2     ; Update X offset
 mov [LineAddressOffsetY],edx

 mov al,[1+ecx]
 test al,0x20       ; Check tile priority
.priority:          ; If pri is 1, jz should be here!
 jz  .check_count   ; dummy jump

 push esi
 and al,0x1C        ; Get palette
 shl al,2
 mov esi,[ecx]      ; Get tile information
 or al,0x0F
 mov ebp,esi        ; Get flip flags
 shl esi,6          ; 8*8
;mov [ColourBase],al
 mov dh,al

 shr ebp,12
 mov eax,[LineAddressOffsetY]
 test ebp,8         ; Check Y flip
 jnz .flip_y
 mov eax,[LineAddressOffset]
.flip_y:
 add esi,eax
 and ebp,byte 4     ; Get X flip

 call dword [Tile_Line_8_4+ebp]
 pop esi

.check_count:
 add edi,byte 8
 pop ecx
 pop ebx

 add ebx,byte 2
;and ebx,byte 31*2
 dec ch
 jnz near .next_tile

 ret

%macro Plot_8_PaletteMask 0
 mov dl,[ColourBase]
;mov ebp,8
 mov ebp,2
%%loop:
 mov eax,[esi]
 add esi,byte 4
 and al,dl
 jz %%no_plot_0
 mov [edi],al
%%no_plot_0:
 add edi,ebx
;dec ebp
;jz %%end
 and ah,dl
 jz %%no_plot_1
 mov [edi],ah
%%no_plot_1:
 add edi,ebx
;dec ebp
;jz %%end
 shr eax,16
 and al,dl
 jz %%no_plot_2
 mov [edi],al
%%no_plot_2:
 add edi,ebx
;dec ebp
;jz %%end
 and ah,dl
 jz %%no_plot_3
 mov [edi],ah
%%no_plot_3:
 add edi,ebx
 dec ebp
 jnz %%loop
%%end:
%endmacro

; New for v0.15, Mode 0 has its own renderers
ALIGNC
LINE_8x8_M0:
.next_tile:
 mov al,[ebx+1]
 test al,0x20       ; Check tile priority
.priority:          ; If pri is 1, jz should be here!
 jz .check_count    ; dummy jump

 mov esi,[ebx]      ; Get tile information
 and al,0x1C        ; Get palette
 mov ebp,esi        ; Get flip flags
 or al,cl           ; Adjust palette for mode 0
 shl esi,6          ; 8*8
;mov [ColourBase],al
 mov dh,al

 shr ebp,12
 mov eax,[LineAddressY]
 test ebp,8         ; Check Y flip
 jnz .flip_y
 mov eax,[LineAddress]
.flip_y:
 add esi,eax
 and ebp,byte 4     ; Get X flip

 call dword [Tile_Line_8_2+ebp]

.check_count:
 add edi,byte 8
 add ebx,byte 2     ; Update screen pointer
 dec ch
 jnz .next_tile
 ret

ALIGNC
; Relies on H-scroll, tile size, tilemap size
EXPORT RENDER_M0_8x8
 mov cl,[M0_Color+edx]
 mov al,[Tile_priority_opcode]
 mov [LINE_8x8_M0.priority],al
 SORT_TILES_8_TALL      ; This sets up the screen vert 2!
 mov edi,[ScreenDestPtr]
 mov eax,[SetAddress+edx]
 add eax,[_TileCache2]
 mov [TilesetAddress],eax

 mov ch,[TileCount1+edx]
 push ebx
 mov ah,[TileCount2+edx]
 push edx
 push eax
 add ebx,[VLMapAddress+edx]
 sub edi,[FirstPixel+edx]
 call LINE_8x8_M0
 mov ch,[1+esp]
 pop esi
 pop edx
 test ch,ch
 pop ebx
 jz .no_second_set
 add ebx,[VRMapAddress+edx]
;call LINE_8x8_M0
 jmp LINE_8x8_M0
.no_second_set:
 ret

ALIGNC
LINE_16x16_M0:
.next_tile:
 mov al,[ebx+1]
 test al,0x20       ; Check tile priority
.priority:          ; If pri is 1, jz should be here!
 jz .check_count    ; dummy jump

 mov esi,[ebx]      ; Get tile information
 and al,0x1C        ; Get palette
 mov ebp,esi        ; Get flip flags
 or al,cl           ; Adjust palette for mode 0
 shl esi,6          ; 8*8
;mov [ColourBase],al
 mov dh,al

 shr ebp,12
 mov eax,[LineAddressY]
 test ebp,8         ; Check Y flip
 jnz .flip_y
 mov eax,[LineAddress]
.flip_y:
 add esi,eax
 and ebp,byte 4     ; Get X flip

 call dword [Tile_Line_16_2+ebp]

.check_count:
 add edi,byte 16
 add ebx,byte 2     ; Update screen pointer
 dec ch
 jnz .next_tile
 ret

ALIGNC
EXPORT RENDER_M0_16x16
 mov cl,[M0_Color+edx]
 mov al,[Tile_priority_opcode]
 mov [LINE_16x16_M0.priority],al
 SORT_TILES_16_TALL     ; This sets up the screen vert 2!
 mov edi,[ScreenDestPtr]
 mov eax,[SetAddress+edx]
 add eax,[_TileCache2]
 mov [TilesetAddress],eax

 mov ch,[TileCount1+edx]
 push ebx
 mov ah,[TileCount2+edx]
 push edx
 push eax
 add ebx,[VLMapAddress+edx]
 sub edi,[FirstPixel+edx]
 call LINE_16x16_M0
 mov ch,[1+esp]
 pop esi
 pop edx
 test ch,ch
 pop ebx
 jz .no_second_set
 add ebx,[VRMapAddress+edx]
;call LINE_16x16_M0
 jmp LINE_16x16_M0
.no_second_set:
 ret

LINE_8x8_C2_Return:
 ret
ALIGNC
LINE_8x8_C2_Check:
 add edi,byte 8
 add ebx,byte 2     ; Update screen pointer
 dec ch
 jz LINE_8x8_C2_Return

LINE_8x8_C2:
 mov al,[ebx+1]
 test al,0x20       ; Check tile priority
.priority:          ; If pri is 1, jz should be here!
 jz LINE_8x8_C2_Check   ; dummy jump

 mov esi,[ebx]      ; Get tile information
 and al,0x1C        ; Get palette
 mov ebp,esi        ; Get flip flags
 or al,3
 shl esi,6          ; 8*8
;mov [ColourBase],al
 mov dh,al

 shr ebp,12
 mov eax,[LineAddressY]
 test ebp,8         ; Check Y flip
 jnz .flip_y
 mov eax,[LineAddress]
.flip_y:
 add esi,eax
 and ebp,byte 4     ; Get X flip

 push dword LINE_8x8_C2_Check
 jmp dword [Tile_Line_8_2+ebp]

LINE_8x8_C4_Return:
 ret
ALIGNC
LINE_8x8_C4_Check:
 add edi,byte 8
 add ebx,byte 2     ; Update screen pointer
 dec ch
 jz LINE_8x8_C4_Return

LINE_8x8_C4:
 mov al,[ebx+1]
 test al,0x20       ; Check tile priority
.priority:          ; If pri is 1, jz should be here!
 jz LINE_8x8_C4_Check   ; dummy jump

 and al,0x1C        ; Get palette

 shl al,2
 mov esi,[ebx]      ; Get tile information
 or al,0x0F
 mov ebp,esi        ; Get flip flags
 shl esi,6          ; 8*8
;mov [ColourBase],al
 mov dh,al

 shr ebp,12
 mov eax,[LineAddressY]
 test ebp,8         ; Check Y flip
 jnz .flip_y
 mov eax,[LineAddress]
.flip_y:
 add esi,eax
 and ebp,byte 4     ; Get X flip

 push dword LINE_8x8_C4_Check
 jmp dword [Tile_Line_8_4+ebp]

LINE_8x8_C8_Return:
 ret
ALIGNC
LINE_8x8_C8_Check:
 add edi,byte 8
 add ebx,byte 2     ; Update screen pointer
 dec ch
 jz LINE_8x8_C8_Return

LINE_8x8_C8:
 mov al,[ebx+1]
 test al,0x20       ; Check tile priority
.priority:          ; If pri is 1, jz should be here!
 jz LINE_8x8_C8_Check   ; dummy jump

 mov esi,[ebx]      ; Get tile information
 mov ebp,esi        ; Get flip flags
 shl esi,6          ; 8*8

 shr ebp,12
 mov eax,[LineAddressY]
 test ebp,8         ; Check Y flip
 jnz .flip_y
 mov eax,[LineAddress]
.flip_y:
 add esi,eax
 and ebp,byte 4     ; Get X flip

 push dword LINE_8x8_C8_Check
 jmp dword [Tile_Line_8_8+ebp]

%macro RENDER_8x8 1
 mov al,[Tile_priority_opcode]
 mov [LINE_8x8_C%1.priority],al
 SORT_TILES_8_TALL      ; This sets up the screen vert 2!
 mov edi,[ScreenDestPtr]
 mov eax,[SetAddress+edx]
 add eax,[_TileCache%1]
 mov [TilesetAddress],eax

 mov ch,[TileCount1+edx]
 push ebx
 mov ah,[TileCount2+edx]
 push edx
 push eax
 add ebx,[VLMapAddress+edx]
 sub edi,[FirstPixel+edx]
 call LINE_8x8_C%1
 mov ch,[1+esp]
 pop esi
 pop edx
 test ch,ch
 pop ebx
 jz .no_second_set
 add ebx,[VRMapAddress+edx]
;call LINE_8x8_C%1
 jmp LINE_8x8_C%1
.no_second_set:
 ret
%endmacro

ALIGNC
LINE_16_Even_C2:
.next_tile:
 mov al,[ebx+1]
 test al,0x20       ; Check tile priority
.priority:          ; If pri is 1, jz should be here!
 jz .check_count    ; dummy jump

 mov esi,[ebx]      ; Get tile information
 and al,0x1C        ; Get palette
 mov ebp,esi        ; Get flip flags
 or al,3            ; Adjust palette for mode 0
 shl esi,6          ; 8*8
;mov [ColourBase],al
 mov dh,al

 shr ebp,12
 mov eax,[LineAddressY]
 test ebp,8         ; Check Y flip
 jnz .flip_y
 mov eax,[LineAddress]
.flip_y:
 add esi,eax
 and ebp,byte 4     ; Get X flip

 call dword [Tile_Line_16_2_Even+ebp]

.check_count:
 add edi,byte 8
 add ebx,byte 2     ; Update screen pointer
 dec ch
 jnz .next_tile
 ret

ALIGNC
LINE_16_Even_C4:
.next_tile:
 mov al,[ebx+1]
 test al,0x20       ; Check tile priority
.priority:          ; If pri is 1, jz should be here!
 jz .check_count    ; dummy jump

 and al,0x1C        ; Get palette
 shl al,2
 mov esi,[ebx]      ; Get tile information
 or al,0x0F
 mov ebp,esi        ; Get flip flags
 shl esi,6          ; 8*8
;mov [ColourBase],al
 mov dh,al

 shr ebp,12
 mov eax,[LineAddressY]
 test ebp,8         ; Check Y flip
 jnz .flip_y
 mov eax,[LineAddress]
.flip_y:
 add esi,eax
 and ebp,byte 4     ; Get X flip

 call dword [Tile_Line_16_4_Even+ebp]

.check_count:
 add edi,byte 8
 add ebx,byte 2     ; Update screen pointer
 dec ch
 jnz .next_tile
 ret

%macro RENDER_16_Even 1
 mov al,[Tile_priority_opcode]
 mov [LINE_16_Even_C%1.priority],al

 mov edi,[ScreenDestPtr]
 mov eax,[SetAddress+edx]
 add eax,[_TileCache%1]
 mov [TilesetAddress],eax

 mov ch,[TileCount1+edx]
 push ebx
 mov ah,[TileCount2+edx]
 push edx
 push eax
 add ebx,[VLMapAddress+edx]
 sub edi,[FirstPixel+edx]
 call LINE_16_Even_C%1
 mov ch,[1+esp]
 pop esi
 pop edx
 test ch,ch
 pop ebx
 jz .no_second_set
 add ebx,[VRMapAddress+edx]
;call LINE_16_Even_C%1
 jmp LINE_16_Even_C%1
.no_second_set:
 ret
%endmacro

ALIGNC
LINE_16x16_C2:
.next_tile:
 mov al,[ebx+1]
 test al,0x20       ; Check tile priority
.priority:          ; If pri is 1, jz should be here!
 jz .check_count    ; dummy jump

 mov esi,[ebx]      ; Get tile information
 and al,0x1C        ; Get palette
 mov ebp,esi        ; Get flip flags
 or al,3            ; Adjust palette for mode 0
 shl esi,6          ; 8*8
;mov [ColourBase],al
 mov dh,al

 shr ebp,12
 mov eax,[LineAddressY]
 test ebp,8         ; Check Y flip
 jnz .flip_y
 mov eax,[LineAddress]
.flip_y:
 add esi,eax
 and ebp,byte 4     ; Get X flip

 call dword [Tile_Line_16_2+ebp]

.check_count:
 add edi,byte 16
 add ebx,byte 2     ; Update screen pointer
 dec ch
 jnz .next_tile
 ret

ALIGNC
LINE_16x16_C4:
.next_tile:
 mov al,[ebx+1]
 test al,0x20       ; Check tile priority
.priority:          ; If pri is 1, jz should be here!
 jz .check_count    ; dummy jump

 and al,0x1C        ; Get palette
 shl al,2
 mov esi,[ebx]      ; Get tile information
 or al,0x0F
 mov ebp,esi        ; Get flip flags
 shl esi,6          ; 8*8
;mov [ColourBase],al
 mov dh,al

 shr ebp,12
 mov eax,[LineAddressY]
 test ebp,8         ; Check Y flip
 jnz .flip_y
 mov eax,[LineAddress]
.flip_y:
 add esi,eax
 and ebp,byte 4     ; Get X flip

 call dword [Tile_Line_16_4+ebp]

.check_count:
 add edi,byte 16
 add ebx,byte 2     ; Update screen pointer
 dec ch
 jnz .next_tile
 ret

ALIGNC
LINE_16x16_C8:
.next_tile:
 mov al,[ebx+1]
 test al,0x20       ; Check tile priority
.priority:          ; If pri is 1, jz should be here!
 jz .check_count    ; dummy jump

 mov esi,[ebx]      ; Get tile information
 mov ebp,esi        ; Get flip flags
 shl esi,6          ; 8*8

 shr ebp,12
 mov eax,[LineAddressY]
 test ebp,8         ; Check Y flip
 jnz .flip_y
 mov eax,[LineAddress]
.flip_y:
 add esi,eax
 and ebp,byte 4     ; Get X flip

 call dword [Tile_Line_16_8+ebp]

.check_count:
 add edi,byte 16
 add ebx,byte 2     ; Update screen pointer
 dec ch
 jnz .next_tile
 ret

%macro RENDER_16x16 1
 mov al,[Tile_priority_opcode]
 mov [LINE_16x16_C%1.priority],al
 SORT_TILES_16_TALL     ; This sets up the screen vert 2!
 mov edi,[ScreenDestPtr]
 mov eax,[SetAddress+edx]
 add eax,[_TileCache%1]
 mov [TilesetAddress],eax

 mov ch,[TileCount1+edx]
 push ebx
 mov ah,[TileCount2+edx]
 push edx
 push eax
 add ebx,[VLMapAddress+edx]
 sub edi,[FirstPixel+edx]
 call LINE_16x16_C%1
 mov ch,[1+esp]
 pop esi
 pop edx
 test ch,ch
 pop ebx
 jz .no_second_set
 add ebx,[VRMapAddress+edx]
;call LINE_16x16_C%1
 jmp LINE_16x16_C%1
.no_second_set:
 ret
%endmacro

ALIGNC
EXPORT RENDER_8x8_C2
    RENDER_8x8 2
ALIGNC
EXPORT RENDER_16x16_C2
%ifdef NO_16x16_TILES
    ret
%endif ; NO_16x16_TILES
    RENDER_16x16 2
ALIGNC
EXPORT RENDER_16x8_Even_C2
    SORT_TILES_8_TALL   ; This sets up the screen vert 2!
    RENDER_16_Even 2
ALIGNC
EXPORT RENDER_16x16_Even_C2
%ifdef NO_16x16_TILES
    ret
%endif ; NO_16x16_TILES
    SORT_TILES_16_TALL  ; This sets up the screen vert 2!
    RENDER_16_Even 2
ALIGNC
EXPORT RENDER_8x8_C4
    RENDER_8x8 4
ALIGNC
EXPORT RENDER_16x16_C4
%ifdef NO_16x16_TILES
    ret
%endif ; NO_16x16_TILES
    RENDER_16x16 4
ALIGNC
EXPORT RENDER_16x8_Even_C4
    SORT_TILES_8_TALL   ; This sets up the screen vert 2!
    RENDER_16_Even 4
ALIGNC
EXPORT RENDER_16x16_Even_C4
%ifdef NO_16x16_TILES
    ret
%endif ; NO_16x16_TILES
    SORT_TILES_16_TALL  ; This sets up the screen vert 2!
    RENDER_16_Even 4
ALIGNC
EXPORT RENDER_8x8_C8
    RENDER_8x8 8
ALIGNC
EXPORT RENDER_16x16_C8
%ifdef NO_16x16_TILES
    ret
%endif ; NO_16x16_TILES
    RENDER_16x16 8

ALIGNC
Sort_Screen_Height:
 ; Corrupts eax,ecx,ebp
 mov al,[HScroll+1+edx]
 mov ecx,TLMapAddress
 test al,1
EXPORT_EQU Tile_Width,$-1   ; 1 = 8x8, 2 = 16x8, 16x16
 mov ebp,TRMapAddress
 jz .first_tile_in_left_screen_map
 add ecx,byte (TRMapAddress-TLMapAddress)
 add ebp,byte (TLMapAddress-TRMapAddress)
.first_tile_in_left_screen_map:
 mov al,[Mosaic+edx]
 test al,al
 mov eax,[Current_Line_Render]
 jz .no_mosaic
 mov eax,[Current_Line_Mosaic]
.no_mosaic:
 add eax,[VScroll+edx]
 test ah,1
EXPORT_EQU Tile_Height,$-1  ; 1 = 8x8, 16x8, 2 = 16x16
 jz .line_in_top_screen_map
 add ecx,byte (BLMapAddress-TLMapAddress)
 add ebp,byte (BLMapAddress-TLMapAddress)
.line_in_top_screen_map:
 mov eax,[edx+ecx]
 add eax,[FirstTile+edx]
 mov [VLMapAddress+edx],eax
 mov eax,[edx+ebp]
 mov [VRMapAddress+edx],eax
 ret

ALIGNC
EXPORT NO_PLANES
 ALL_SPRITES_PLANE 3
 ALL_SPRITES_PLANE 2
 ALL_SPRITES_PLANE 1
 ALL_SPRITES_PLANE 0
NO_PLANES_S:
 ret

;%1 = planenum, %2 = pri
%macro RENDER_LINE 2
  LOAD_BG_TABLE %1
  mov byte [Tile_priority_opcode],0x75-%2
  ; tile size and depth selected in BGMODE write handler
  call dword [LineRender+edx]
%endmacro

%if 0
 Mode 0 - 4 background layers, 8x8 and 16x16 tile sizes
  2bpl tile depth in all layers
  special: BG3 high priority selection
           each background layer has its own set of palettes

 Mode 1 - 3 background layers, 8x8 and 16x16 tile sizes
  4bpl tile depth in layers 1 and 2, 2bpl tile depth in layer 3
  special: BG3 high priority selection
%endif
ALIGNC
EXPORT_C SCREEN_MODE_0
EXPORT SCREEN_MODE_0
EXPORT_C SCREEN_MODE_1
EXPORT SCREEN_MODE_1
; There are 15 Versions of Render & Merge functions (4 planes)

 push eax

 test al,4
 jz .no_bg3_max
 test byte [_BGMODE],8
 jz .no_bg3_max
 RENDER_LINE 3,1
.no_bg3_max:

 test byte [esp],0x10
 jz .no_spr_3
 ALL_SPRITES_PLANE 3
.no_spr_3:

 test byte [esp],1
 jz .no_bg1_hi
 RENDER_LINE 1,1
.no_bg1_hi:

 test byte [esp],2
 jz .no_bg2_hi
 RENDER_LINE 2,1
.no_bg2_hi:

 test byte [esp],0x10
 jz .no_spr_2
 ALL_SPRITES_PLANE 2
.no_spr_2:

 test byte [esp],1
 jz .no_bg1_lo
 RENDER_LINE 1,0
.no_bg1_lo:

 test byte [esp],2
 jz .no_bg2_lo
 RENDER_LINE 2,0
.no_bg2_lo:

 test byte [esp],0x10
 jz .no_spr_1
 ALL_SPRITES_PLANE 1
.no_spr_1:

 test byte [esp],4
 jz .no_bg3_hi
 test byte [_BGMODE],8
 jnz .no_bg3_hi
 RENDER_LINE 3,1
.no_bg3_hi:

 test byte [esp],8
 jz .no_bg4_hi
 RENDER_LINE 4,1
.no_bg4_hi:

 test byte [esp],0x10
 jz .no_spr_0
 ALL_SPRITES_PLANE 0
.no_spr_0:

 test byte [esp],4
 jz .no_bg3_lo
 RENDER_LINE 3,0
.no_bg3_lo:

 test byte [esp],8
 jz .no_bg4_lo
 RENDER_LINE 4,0
.no_bg4_lo:

 pop eax
 ret

%if 0
 Mode 2 - 2 background layers, 8x8 and 16x16 tile sizes
  4bpl tile depth in layers 1 and 2
  special: offset change data (h and v) stored in layer 3

 Mode 3 - 2 background layers, 8x8 and 16x16 tile sizes
  8bpl tile depth in layer 1, 4bpl tile depth in layer 2
  special: direct color mode

 Mode 4 - 2 background layers, 8x8 and 16x16 tile sizes
  8bpl tile depth in layer 1, 2bpl tile depth in layer 2
  special: direct color mode
           offset change data (h or v?) stored in layer 3

 Mode 5 - 2 background layers, 16x8 and 16x16 tile sizes
  4bpl tile depth in layer 1, 2bpl tile depth in layer 2
  special: 512 mode

 Mode 6 - 1 background layer, 16x8 and 16x16 tile sizes
  4bpl tile depth in layer 1
  special: 512 mode
           offset change data (h and v?) stored in layer 3 (?)
%endif
EXPORT_C SCREEN_MODE_2
EXPORT SCREEN_MODE_2
EXPORT_C SCREEN_MODE_3
EXPORT SCREEN_MODE_3
EXPORT_C SCREEN_MODE_4
EXPORT SCREEN_MODE_4
EXPORT_C SCREEN_MODE_5
EXPORT SCREEN_MODE_5
EXPORT_C SCREEN_MODE_6
EXPORT SCREEN_MODE_6
; There are 3 Versions of Render & Merge functions (2 planes)

 push eax

 test byte [esp],0x10
 jz .no_spr_3
 ALL_SPRITES_PLANE 3
.no_spr_3:

 test byte [esp],1
 jz .no_bg1_hi
 RENDER_LINE 1,1
.no_bg1_hi:

 test byte [esp],0x10
 jz .no_spr_2
 ALL_SPRITES_PLANE 2
.no_spr_2:

 test byte [esp],2
 jz .no_bg2_hi
 RENDER_LINE 2,1
.no_bg2_hi:

 test byte [esp],0x10
 jz .no_spr_1
 ALL_SPRITES_PLANE 1
.no_spr_1:

 test byte [esp],1
 jz .no_bg1_lo
 RENDER_LINE 1,0
.no_bg1_lo:

 test byte [esp],0x10
 jz .no_spr_0
 ALL_SPRITES_PLANE 0
.no_spr_0:

 test byte [esp],2
 jz .no_bg2_lo
 RENDER_LINE 2,0
.no_bg2_lo:

 pop eax
 ret

section .text
ALIGNC
section .data
ALIGND
section .bss
ALIGNB
